/*
**  linux/amiga/amicon_ocs.c
**
**  Copyright (C) 1994 David Carter  [carter@compsci.bristol.ac.uk]
**      Modified 20-Apr-94 by William Rucklidge (wjr@cs.cornell.edu)
**          - AGA modes
**      Modified 03-May-94 by Geert Uytterhoeven (uytterho@cs.kuleuven.ac.be)
**          - additional AGA modes         
**          - 24-bit color palette
**      Modified 19-June-94 by Geert Uytterhoeven
**          - cursor flashing for better cursor visibility
**
**  Interface based on the original 'amicon.c' code:
**      Copyright (C) 1993 Hamish Macdonald and Greg Harp
**  although the contents have been radically altered.
**
** This file is subject to the terms and conditions of the GNU General
** Public License.  See the file README.legal in the main directory of this
** archive for more details.
*/

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>

#include <linux/amigahw.h>
#include <linux/config.h>       /* New: DPC */
#include <linux/bootinfo.h>
#include <linux/interrupt.h>    /* New: DPC */
#include <linux/amigaints.h>

#include <asm/system.h>

/*
 *  Color definitions
 */

#define FG_COLOR	(0x000000)	/* black */
#define BG_COLOR	(0xaaaaaa)	/* lt. grey */
#define CRSR_COLOR	(0xff0000)	/* bright red */

#define FG_COLOR_INV	BG_COLOR
#define BG_COLOR_INV	FG_COLOR
#define CRSR_COLOR_INV	(0x4477aa)

/*
 *  Split 24-bit RGB colors in 12-bit MSB (for OCS/ECS/AGA) and LSB (for AGA)
 */

#define COLOR_MSB(rgb)	(((rgb>>12)&0xf00)|((rgb>>8)&0x0f0)|((rgb>>4)&0x00f))
#define COLOR_LSB(rgb)	(((rgb>>8) &0xf00)|((rgb>>4)&0x0f0)|((rgb)   &0x00f))

/* Cursor definitions */

#define CRSR_FLASH	1	/* Cursor flashing on(1)/off(0) */
#define CRSR_RATE	24	/* Number of frames/flash toggle */
#define CRSR_BLOCK	1	/* Block(1) or line(0) cursor */

/*
 * A bunch of register definitions. Doing some things in terms of bitfields
 * might be nice eventually.
 */

/* BPLCON0 defines */
#define	BPC0_HIRES	0x8000
#define	BPC0_BPU3	0x0010	/* AGA */
#define	BPC0_BPU2	0x4000	/* Bit plane used count */
#define	BPC0_BPU1	0x2000
#define	BPC0_BPU0	0x1000
#define	BPC0_HAM	0x0800	/* HAM mode */
#define	BPC0_DPF	0x0400	/* Double playfield */
#define	BPC0_COLOR	0x0200	/* Enable colorburst */
#define	BPC0_GAUD	0x0100	/* Genlock audio enable */
#define	BPC0_UHRES	0x0080	/* Ultrahi res enable */
#define	BPC0_SHRES	0x0040	/* Super hi res mode */
#define	BPC0_BYPASS	0x0020	/* Bypass LUT - AGA */
#define	BPC0_LPEN	0x0008	/* Light pen enable */
#define	BPC0_LACE	0x0004	/* Interlace */
#define	BPC0_ERSY	0x0002	/* External resync */
#define	BPC0_ECSENA	0x0001	/* ECS emulation disable */

/* BPLCON1 controlls horizontal scrolling */

#define	BPC2_KILLEHB	0x0200	/* Kill EHB mode - AGA */
#define	BPC2_RDRAM	0x0100	/* Color table accesses read, not write - AGA */
#define	BPC2_SOGEN	0x0080	/* SOG output pin high - AGA */
#define	BPC2_PF2PRI	0x0040	/* PF2 priority over PF1 */
#define	BPC2_PF2P2	0x0020	/* PF2 priority wrt sprites */
#define	BPC2_PF2P1	0x0010
#define	BPC2_PF2P0	0x0008
#define	BPC2_PF1P2	0x0004	/* ditto PF1 */
#define	BPC2_PF1P1	0x0002
#define	BPC2_PF1P0	0x0001

/* BPLCON3 defines - AGA */
#define	BPC3_BANK2	0x8000	/* Bits to select color register bank */
#define	BPC3_BANK1	0x4000
#define	BPC3_BANK0	0x2000
#define	BPC3_PF2OF2	0x1000	/* Bits for color table offset when PF2 */
#define	BPC3_PF2OF1	0x0800
#define	BPC3_PF2OF0	0x0400
#define	BPC3_LOCT	0x0200	/* Color register writes go to low bits */
#define	BPC3_SPRES1	0x0080	/* Sprite resolution bits */
#define	BPC3_SPRES0	0x0040
#define	BPC3_BRDRBLNK	0x0020	/* Border blanked? */
#define	BPC3_BRDRTRAN	0x0010	/* Border transparent? */
#define	BPC3_BRDRSPRT	0x0002	/* Sprites in border? */
#define	BPC3_EXTBLKEN	0x0001	/* BLANK programmable */

/* BEAMCON0 */
#define	BMC0_HARDDIS	0x4000	/* Disable hardware limits */
#define	BMC0_LPENDIS	0x2000	/* Disable light pen latch */
#define	BMC0_VARVBEN	0x1000	/* Enable variable vertical blank */
#define	BMC0_LOLDIS	0x0800	/* Disable long/short line toggle */
#define	BMC0_CSCBEN	0x0400	/* Composite sync/blank */
#define	BMC0_VARVSYEN	0x0200	/* Enable variable vertical sync */
#define	BMC0_VARHSYEN	0x0100	/* Enable variable horizontal sync? */
#define	BMC0_VARBEAMEN	0x0080	/* Enable variable beam counters */
#define	BMC0_PAL	0x0020	/* Set decodes for PAL? */
#define	BMC0_VARCSYEN	0x0010	/* Enable variable composite sync? */
#define	BMC0_BLANKEN	0x0008	/* Blank enable? */
#define	BMC0_CSYTRUE	0x0004	/* CSY polarity */
#define	BMC0_VSYTRUE	0x0002	/* VSY polarity */
#define	BMC0_HSYTRUE	0x0001	/* HSY polarity */

/* FMODE - AGA */
#define	FMODE_SSCAN2	0x8000	/* Sprite scan-doubling */
#define	FMODE_BSCAN2	0x4000	/* Use PF2 modulus every other line */
#define	FMODE_SPAGEM	0x0008	/* Sprite page mode */
#define	FMODE_SPR32	0x0004	/* Sprite 32 bit fetch */
#define	FMODE_BPAGEM	0x0002	/* Bitplane page mode */
#define	FMODE_BPL32	0x0001	/* Bitplane 32 bit fetch */


#define CUSTOM_OFS(fld) ((long)&((struct CUSTOM*)0)->fld)

/*
** Externals
*/

extern unsigned char
    fontdata[];

extern int
    fontheight,
    fontwidth;

extern char
    fontname[];

/*
** Local Types
*/

struct display
{
    ushort scr_max_height;      /* screen dimensions */
    ushort scr_max_width;
    ushort scr_height;
    ushort scr_width;
    ushort scr_depth;
    ushort bytes_per_row;

    ushort scroll_latch;        /* Vblank support for hardware scroll */
    ushort y_wrap;      
    ushort cursor_latch;        /* Hardware cursor */
    ushort cursor_x;            
    ushort cursor_y;
    ushort *cursor, *dummy;
    ushort cursor_flash;

    /* Some chipreg values we need to rebuild copper lists */
    ushort diwstrt_v, diwstrt_h;   /* display window control */
    ushort diwstop_v, diwstop_h;
    ushort bplcon0;                /* display mode */

    ulong fgcol;               /* 24-bit text colors */
    ulong bgcol;
    ulong crsrcol;

    u_char *bitplane[8];        /* pointers to display bitplanes */
    ulong   plane_size;
    u_long *color_map;          /* XXX Not yet implemented XXX */

    u_char *fontdata;           /* Font associated to this display */
    ushort fontheight;
    ushort fontwidth;

    ushort *coplist1hdr;   /* List 1 static  component */
    ushort *coplist1dyn;   /* List 1 dynamic component */
    ushort *coplist2hdr;   /* List 2 static  component */
    ushort *coplist2dyn;   /* List 2 dynamic component */
    };


/* Geometry structure contains all useful information about given mode.
 *
 * Strictly speaking `scr_max_height' and `scr_max_width' is redundant
 * information with DIWSTRT value provided. Might be useful if modes
 * can be hotwired by user in future. It fits for the moment.
 *
 * At the moment, the code only distinguishes between OCS and AGA. ECS
 * lies somewhere in between - someone more familiar with it could make
 * appropriate modifications so that some advanced display modes are
 * available, without confusing the poor chipset. OCS modes use only the
 * bplcon0, diwstrt, diwstop, ddfstrt, ddfstop registers (a few others could
 * be used as well). -wjr
 */
 
struct geometry {
    char *modename;	/* Name this thing */
    char isOCS;		/* Is it OCS or AGA */
    ushort bplcon0;	/* Values for bit plane control register 0 */
    ushort scr_width;
    ushort scr_height;
    ushort scr_depth;
    ushort scr_max_width;     /* Historical, might be useful still */
    ushort scr_max_height;
    ushort diwstrt_h;	/* Where the display window starts */
    ushort diwstrt_v;
    ushort alignment;	/* Pixels per scanline must be a multiple of this */
    /* OCS doesn't need anything past here */
    ushort bplcon2;
    ushort bplcon3;
    /* The rest of these control variable sync */
    ushort htotal;	/* Total hclocks */
    ushort hsstrt;	/* HSYNC start and stop */
    ushort hsstop;
    ushort hbstrt;	/* HBLANK start and stop */
    ushort hbstop;
    ushort vtotal;	/* Total vlines */
    ushort vsstrt;	/* VSYNC, VBLANK ditto */
    ushort vsstop;
    ushort vbstrt;
    ushort vbstop;
    ushort hcenter;	/* Center of line, for interlaced modes */
    ushort beamcon0;	/* Beam control */
    ushort fmode;	/* Memory fetch mode */
};

#define MAX_COP_LIST_ENTS 64
#define COP_MEM_REQ       (MAX_COP_LIST_ENTS*4*2)
#define SPR_MEM_REQ       (24)

/*
** Local Variables
*/

static struct display disp[NR_CONSOLES];

static ushort cursor_data[] =
{
    0x2c81,0x2d00,
    0xf000,0x0000,
    0x0000,0x0000
};

/* controlling screen blanking (read in VBL handler) */
static int do_blank;
static unsigned short save_bplcon3;
static int do_unblank;

/*
** Functions
*/

static struct {
   int    right_count;
   int    done;
} vblank;

/* Couple of little support routines for debugging */

void wait_for_vblank(void) 
{
   vblank.done = 0;

   while (!vblank.done)
      ;
}

int right_count(void)
{
   return(vblank.right_count);
}


void update_y_wrap(int unit, int y_wrap)
{
   register struct display *p = &disp[unit];

   p->scroll_latch = 0;
   p->y_wrap   = (ushort) y_wrap;
   p->scroll_latch = 1;
}

static __inline__ void init_vblank(void)
{
   vblank.right_count  = 0;
   vblank.done         = 0;
}

/* Notes about copper scrolling:
 *
 * 1. The VBLANK routine dynamically rewrites a LIVE copper list that is
 *    currently being executed. Don't mess with it unless you know the
 *    complications. Fairly sure that double buffered lists doesn't
 *    make our life any easier.
 *
 * 2. The vblank code starts executing at logical line 0. Display must be
 *    set up and ready to run by line DIWSTRT_V, typically 0x2c, minimum
 *    value is 0x18 for maximum overscan.
 *
 *    Tests on my A500/030 for dynamically generating a 37 element copper
 *    list during the VBLANK period under AmigaDos required between
 *    0x10 and 0x14 scanlines. This should be pathological case, and
 *    should do better under Linux/68k. It is however IMPERATIVE that I am
 *    first in the VBLANK isr chain. Try to keep 'buildclist' as fast as
 *    possible. Don't think that it justifies assembler thou'
 *
 * 3. PAL 640x256 display (no overscan) has copper-wait y positions in range
 *    0x02c -> 0x12c. NTSC overscan uses values > 256 too. However counter
 *    is 8 bit, will wrap. RKM 1.1 suggests use of a WAIT(0x00,0xff),
 *    WAIT(x,y-0x100) pair to handle this case. This is WRONG - must use
 *    WAIT(0xe2,0xff) to ensure that wrap occures by next copper
 *    instruction. Argghh!
 *
 * 4. RKM 1.1 suggests Copper-wait x positions are in range [0,0xe2].
 *    Horizontal blanking occurs in range 0x0f -> 0x35. Black screen
 *    shown in range 0x04 -> 0x47.
 *
 *    Experiments suggest that using WAIT(0x00,y), we can replace up to
 *    7 bitplane pointers before display fetch start. Using a 
 *    WAIT(0xe0,y-1) instead, we can replace 8 pointers that should be
 *    all that we need for a full AGA display. Should work because of
 *    fetch latency with bitmapped display.
 *
 *    I think that this works. Someone please tell me if something breaks.
 *
 * Is diwstop_h the right value to use for "close to the end of line"?
 * It seems to work for me, at least for the modes I've defined. -wjr
 *
 * I changed the Wait(diwstop_h, 0xff) for 256-line chunk skipping to
 * Wait(diwstop_h-2, 0xff) to make it work with the additional
 * `get-all-you-can-get-out-of-it' AGA modes. Maybe we should derive the
 * wait position from the HTOTAL value? - G.U.
 */

static __inline__ ushort *
    build_clist_hdr(register struct display *p,
                    ushort *cop,
                    ushort *othercop) /* Interlace: List for next frame */
{
    int    i;
    ushort diwstrt_v = p->diwstrt_v;
    ushort diwstop_h = p->diwstop_h;

    if (othercop) {
        *cop++ = CUSTOM_OFS(cop1lc);
        *cop++ = (long)othercop >> 16;
        *cop++ = CUSTOM_OFS(cop1lc) + 2;
        *cop++ = (long)othercop;
    }

    /* Point Sprite 0 at cursor sprite: */
    *cop++ = CUSTOM_OFS(sprpt[0]);
    *cop++ = (ushort)((long)p->cursor >> 16);
    *cop++ = CUSTOM_OFS(sprpt[0]) + 2;
    *cop++ = (ushort)((long)p->cursor & 0x0000ffff);

    /* Point Sprites 1-7 at dummy sprite: */
    for (i=1; i<8; i++) {
	*cop++ = CUSTOM_OFS(sprpt[i]);
	*cop++ = (ushort)((long)p->dummy >> 16);
	*cop++ = CUSTOM_OFS(sprpt[i]) + 2;
	*cop++ = (ushort)((long)p->dummy & 0x0000ffff);
    }

    /* Halt copper until we have rebuilt the display list */

    *cop++ = ((diwstrt_v - 2)   << 8) | (diwstop_h >> 1) | 0x1;
    *cop++ = 0xfffe;

    return(cop);
}

static __inline__ ushort *
    build_clist_dyn(register struct display *p,
                    ushort *cop,
                    int shf)       /* Interlace: Short frame      */
{
    ushort diwstrt_v = p->diwstrt_v;
    ushort diwstop_h = p->diwstop_h;
    ushort y_wrap    = p->y_wrap;
    ulong offset    = y_wrap * p->bytes_per_row;
    long   scrmem;
    int    i;

    /* Set up initial bitplane ptrs */

    for (i = 0 ; i < p->scr_depth ; i++) {
        scrmem    = ((long)p->bitplane[i]) + offset;
        if (shf)
            scrmem += p->bytes_per_row;

        *cop++ = CUSTOM_OFS(bplpt[i]);
        *cop++ = (long)scrmem >> 16;
        *cop++ = CUSTOM_OFS(bplpt[i]) + 2;
        *cop++ = (long)scrmem;
    }

    /* If wrapped frame needed - wait for line then switch bitplXs */

    if (y_wrap) {
        ushort line;
        
        if (p->bplcon0 & BPC0_LACE)
            line = diwstrt_v + (p->scr_height - y_wrap)/2;
        else
            line = diwstrt_v + p->scr_height - y_wrap;

	/* Handle skipping over 256-line chunks */
        while (line > 256) {
	    /* Hardware limitation - 8 bit counter    */
	    /* Wait(diwstop_h-2, 0xff) */
            *cop++ = 0xff00 | ((diwstop_h-2) >> 1) | 0x1;
            *cop++ = 0xfffe;
	    /* Wait(0, 0) - make sure we're in the new segment */
            *cop++ = 0x0001;
            *cop++ = 0xfffe;
	    line -= 256;
	    }
	
	/* Wait(diwstop_h, line - 1) */
	*cop++ = ((line - 1)   << 8) | (diwstop_h >> 1) | 0x1;
	*cop++ = 0xfffe;

        for (i = 0 ; i < p->scr_depth ; i++) {
            scrmem = (long)p->bitplane[i];
        
            if (shf)
                scrmem += p->bytes_per_row;

            *cop++ = CUSTOM_OFS(bplpt[i]); 
            *cop++ = (long)scrmem >> 16;
            *cop++ = CUSTOM_OFS(bplpt[i]) + 2;
            *cop++ = (long)scrmem;
        }
    }
    
    /* End of Copper list */
    *cop++ = 0xffff;
    *cop++ = 0xfffe;

    return(cop);
}

static __inline__ void build_cursor(register struct display *p)
{
   int vs, hs, ve;
   ushort diwstrt_v = p->diwstrt_v;
   ushort diwstrt_h = p->diwstrt_h;

   if (p->bplcon0 & BPC0_LACE) {
      vs = diwstrt_v + (p->cursor_y * p->fontheight)/2; 
      ve = vs + p->fontheight/2;
   } else {
      vs = diwstrt_v + (p->cursor_y * p->fontheight); 
      ve = vs + p->fontheight;
   }

   if (p->bplcon0 & BPC0_ECSENA) {
       /*
	* It's an AGA mode. We'll assume that the sprite was set
	* into 35ns resolution by the appropriate SPRES bits in bplcon3.
	*/
       hs = diwstrt_h  * 4 + (p->cursor_x * p->fontwidth) - 4;
       }
   else {
       hs = diwstrt_h + (p->cursor_x * p->fontwidth) / 2 - 1;
       }

   if (p->bplcon0 & BPC0_ECSENA) {
       /* There are some high-order bits on the sprite position */
       *((ulong *) p->cursor) =
	   ((((vs & 0xff) << 24) | ((vs & 0x100) >> 6) |
	     ((vs & 0x200) >> 3)) |
	    (((hs & 0x7f8) << 13) | ((hs & 0x4) >> 2) |
	     ((hs & 0x3) << 3)) |
	    (((ve & 0xff) << 8) | ((ve & 0x100) >> 7) |
	     ((ve & 0x200) >> 4)));
       }
   else {
       *((ulong *) p->cursor) =
	   ((vs << 24) | ((vs & 0x00000100) >> 6) |
	    ((hs & 0x000001fe) << 15) | (hs & 0x00000001) | 
	    ((ve & 0x000000ff) << 8) | ((ve & 0x00000100) >> 7));
       }
   }

static void amicon_interrupt(struct intframe *fp, void *data)
{
   int unit = 0;
   register struct display *p = &disp[unit];

   static ushort cursorcount = 0;
   static ushort cursorstate = 0;

   /* I *think* that you should only change display lists on long frame.
    * At least it goes awfully perculiar on my A500 without the following
    * test. Not really in a position to test this hypothesis, so sorry
    * for the slow scrolling, all you flicker-fixed souls
   */

   if (!(p->bplcon0 & BPC0_LACE) || (custom.vposr & 0x8000)) {
       if (p->scroll_latch || p->cursor_latch)
           build_cursor(p);

       if (p->scroll_latch)
           if (p->bplcon0 & BPC0_LACE) {
               build_clist_dyn(p, disp[unit].coplist1dyn, 0);
               build_clist_dyn(p, disp[unit].coplist2dyn, 1);
           } else
               build_clist_dyn(p, disp[unit].coplist1dyn, 0);

       p->scroll_latch = 0;
       p->cursor_latch = 0;
   }

   if (!(custom.potgor & (1<<10)))
       vblank.right_count++;

   if (p->cursor_flash) {
      if (cursorcount)
         cursorcount--;
      else {
         cursorcount = CRSR_RATE;
         if ((cursorstate = !cursorstate))
            custom.dmacon = DMAF_SETCLR | DMAF_SPRITE;
         else
            custom.dmacon = DMAF_SPRITE;
      }
   }
   
   if (do_blank) {
       custom.dmacon = DMAF_RASTER | DMAF_SPRITE;
       custom.color[0]  = 0;
       if (boot_info.bi_amiga.chipset == CS_AGA) {
	   /* Fill in the LSB of the 24 bit color palette */
	   /* Must happen after MSB */
	   custom.bplcon3   = save_bplcon3 | BPC3_LOCT;
	   custom.color[0]  = 0;
	   custom.bplcon3   = save_bplcon3;
       }
       do_blank = 0;
   }

   if (do_unblank) {
       custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_SPRITE;
       custom.color[0]  = COLOR_MSB(p->bgcol);
       if (boot_info.bi_amiga.chipset == CS_AGA) {
	   /* Fill in the LSB of the 24 bit color palette */
	   /* Must happen after MSB */
	   custom.bplcon3   = save_bplcon3 | BPC3_LOCT;
	   custom.color[0]  = COLOR_LSB(p->bgcol);
	   custom.bplcon3   = save_bplcon3;
       }
       do_unblank = 0;
   }

   vblank.done  = 1;
}

/* ====================================================================== */


/* display_init():
 *
 *    Fills out (struct display *) given a geometry structure
*/

static void display_init(struct display *p,
                         struct geometry *geom, ushort inverse)
{
   int i;
   char *chipptr;
   ushort diwstrt_v, diwstop_v;
   ushort diwstrt_h, diwstop_h;
   ushort diw_min_h, diw_min_v;
   ushort bplmod, diwstrt, diwstop, diwhigh, ddfstrt, ddfstop;
   ushort cursorheight, cursormask = 0;

   /* Decide colour scheme */

   if (inverse) {
       p->fgcol   = FG_COLOR_INV;
       p->bgcol   = BG_COLOR_INV;
       p->crsrcol = CRSR_COLOR_INV;
   } else {
       p->fgcol   = FG_COLOR;
       p->bgcol   = BG_COLOR;
       p->crsrcol = CRSR_COLOR;
   }

   /* Define screen geometry */
   
   p->scr_max_height = geom->scr_max_height;
   p->scr_max_width  = geom->scr_max_width; 
   p->scr_height     = geom->scr_height;
   p->scr_width      = geom->scr_width;
   p->scr_depth      = geom->scr_depth;
   p->fontdata       = fontdata;
   p->fontheight     = fontheight;
   p->fontwidth      = fontwidth;
   p->bplcon0        = geom->bplcon0 | BPC0_COLOR;
   if (p->scr_depth < 8) {
       p->bplcon0 |= (p->scr_depth << 12);
       }
   else {
       /* must be exactly 8 */
       p->bplcon0 |= BPC0_BPU3;
       }

   diw_min_v         = geom->diwstrt_v;
   diw_min_h         = geom->diwstrt_h;

   /* We can derive everything else from this, at least for OCS */
   /*
    * For AGA: we don't use the finer position control available for
    * diw* yet (could be set by 35ns increments).
    */

   /* Calculate line and plane size while respecting the alignment restrictions */
   p->bytes_per_row  = ((p->scr_width+geom->alignment-1)&~(geom->alignment-1)) >> 3;
   p->plane_size     = p->bytes_per_row * p->scr_height;

   p->y_wrap   = 0;                  p->scroll_latch = 1;
   p->cursor_x = 0; p->cursor_y = 0; p->cursor_latch = 1;

   if (p->bplcon0 & BPC0_LACE) {
       bplmod = p->bytes_per_row;
       diwstrt_v = diw_min_v + (p->scr_max_height - p->scr_height)/4;
       diwstop_v = (diwstrt_v + p->scr_height/2);
   } else {
       bplmod = 0;
       diwstrt_v = diw_min_v + (p->scr_max_height - p->scr_height)/2;
       diwstop_v = (diwstrt_v + p->scr_height);
   }

   if (p->bplcon0 & BPC0_HIRES) {
       diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/4;
       diwstop_h = (diwstrt_h + p->scr_width/2);
          /* ??? Where did 0x1d5 come from in original code ??? */
       }
   else if (p->bplcon0 & BPC0_SHRES) {
       diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/8;
       diwstop_h = (diwstrt_h + p->scr_width/4);
       }
   else {
       diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/2;
       diwstop_h = (diwstrt_h + p->scr_width);
       }

   if (p->bplcon0 & BPC0_HIRES) {
       ddfstrt = (diwstrt_h >> 1) - 4;
       ddfstop = ddfstrt + (4 * (p->bytes_per_row>>1)) - 8;
       }
   else if (p->bplcon0 & BPC0_SHRES) {
       /* There may be some interaction with FMODE here... -8 is magic. */
       ddfstrt = (diwstrt_h >> 1) - 8;
       ddfstop = ddfstrt + (2 * (p->bytes_per_row>>1)) - 8;
       }
   else {
       ddfstrt = (diwstrt_h >> 1) - 8;
       ddfstop = ddfstrt + (8 * (p->bytes_per_row>>1)) - 8;
       }

   if (p->bplcon0 & BPC0_LACE)
      cursorheight = p->fontheight/2;
   else
      cursorheight = p->fontheight;

   chipptr = amiga_chip_alloc(p->scr_depth*p->plane_size+COP_MEM_REQ+SPR_MEM_REQ+4*(cursorheight-1));
   
   /* locate the bitplanes */
   /* These MUST be 64 bit aligned for full AGA compatibility!! */
   for (i=0 ; i<p->scr_depth ; i++, chipptr += p->plane_size) {
       p->bitplane[i] = (u_char *) chipptr;
       memset ((void *)chipptr, 0, p->plane_size);  /* and clear */
       }

   /* locate the copper lists */
   p->coplist1hdr = (ushort *) chipptr;  chipptr += MAX_COP_LIST_ENTS * 4;
   p->coplist2hdr = (ushort *) chipptr;  chipptr += MAX_COP_LIST_ENTS * 4;

   /* locate the sprite data */
   p->cursor      = (ushort *) chipptr;  chipptr += 8+4*cursorheight;
   p->dummy       = (ushort *) chipptr;  chipptr += 12;

   /* create the sprite data for the cursor image */
   memset((void *)p->cursor, 0, 8+4*cursorheight);
   if (p->bplcon0 & BPC0_ECSENA)
      /* AGA cursor is SHIRES */
      for (i = 0; (i < p->fontwidth) && (i < 16); i++)
         cursormask |= 1<<(15-i);
   else
      /* OCS cursor is LORES */
      for (i = 0; (i < p->fontwidth/2) && (i < 8); i++)
         cursormask |= 1<<(15-i);

   p->cursor[0] = cursor_data[0];
   p->cursor[1] = cursor_data[1];

#if (CRSR_BLOCK == 1)
   for (i = 0; i < cursorheight; i++)
#else
   for (i = cursorheight-2; i < cursorheight; i++)
#endif
      p->cursor[2+2*i] = cursormask;

   /* set dummy sprite data to a blank sprite */
   memset((void *)p->dummy, 0, 12);
  
   /* set cursor flashing */
   p->cursor_flash = CRSR_FLASH;
 
   /* Initialise the chipregs */
   p->diwstrt_v = diwstrt_v;
   p->diwstrt_h = diwstrt_h;
   p->diwstop_v = diwstop_v;
   p->diwstop_h = diwstop_h;
   diwstrt = ((diwstrt_v << 8) | diwstrt_h);
   diwstop = ((diwstop_v & 0xff) << 8) | (diwstop_h & 0xff);

   custom.bplcon0   = p->bplcon0;	/* set the display mode */
   custom.bplcon1   = 0;		/* Needed for horizontal scrolling */
   custom.bplcon2   = 0;
   custom.bpl1mod   = bplmod;
   custom.bpl2mod   = bplmod;
   custom.diwstrt   = diwstrt;
   custom.diwstop   = diwstop;
   custom.ddfstrt   = ddfstrt;
   custom.ddfstop   = ddfstop;

   custom.color[0]  = COLOR_MSB(p->bgcol);
   custom.color[1]  = COLOR_MSB(p->fgcol);
   custom.color[17] = COLOR_MSB(p->crsrcol); /* Sprite 0 color */

   if (boot_info.bi_amiga.chipset == CS_AGA) {
      /* Fill in the LSB of the 24 bit color palette */
      /* Must happen after MSB */
      custom.bplcon3   = geom->bplcon3 | BPC3_LOCT;
      custom.color[0]  = COLOR_LSB(p->bgcol);
      custom.color[1]  = COLOR_LSB(p->fgcol);
      custom.color[17] = COLOR_LSB(p->crsrcol);
      custom.bplcon3   = geom->bplcon3;
   }
   if (!(geom->isOCS)) {
       /* Need to set up a bunch more regs */
       /* Assumes that diwstrt is in the (0,0) sector, but stop might not be */
       diwhigh = (diwstop_v & 0x700) | ((diwstop_h & 0x100) << 5);

       custom.bplcon2   = geom->bplcon2;
       custom.bplcon3   = geom->bplcon3;
       /* save bplcon3 for blanking */
       save_bplcon3 = geom->bplcon3;

       custom.diwhigh   = diwhigh;	/* must happen AFTER diwstrt, stop */

       custom.htotal	= geom->htotal;
       custom.hsstrt	= geom->hsstrt;
       custom.hsstop	= geom->hsstop;
       custom.hbstrt	= geom->hbstrt;
       custom.hbstop	= geom->hbstop;
       custom.vtotal	= geom->vtotal;
       custom.vsstrt	= geom->vsstrt;
       custom.vsstop	= geom->vsstop;
       custom.vbstrt	= geom->vbstrt;
       custom.vbstop	= geom->vbstop;
       custom.hcenter	= geom->hcenter;
       custom.beamcon0	= geom->beamcon0;
       custom.fmode	= geom->fmode;

       /* We could load 8-bit colors here, if we wanted */
       }


   /* Build initial copper lists. sprites must be set up, and p->diwstrt. */

   if (p->bplcon0 & BPC0_LACE) {
       p->coplist1dyn = build_clist_hdr(p,p->coplist1hdr, p->coplist2hdr),
       build_clist_dyn(p, p->coplist1dyn, 0);

       p->coplist2dyn = build_clist_hdr(p,p->coplist2hdr, p->coplist1hdr),
       build_clist_dyn(p, p->coplist2dyn, 1);
       }
   else {
       p->coplist1dyn = build_clist_hdr(p,p->coplist1hdr, NULL),
       build_clist_dyn(p, p->coplist1dyn, 0);
       }


   /* Get ready to run first copper list */
   custom.cop1lc = p->coplist1hdr;
   custom.copjmp1 = 0;

   /* turn on DMA for bitplane and sprites */
   custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_COPPER | DMAF_SPRITE;

   if (p->bplcon0 & BPC0_LACE) {
       /* Make sure we get the fields in the right order */

       /* wait for LOF frame bit to go low */
       while (custom.vposr & 0x8000)
           ;

       /* wait for LOF frame bit to go high */
       while (!(custom.vposr & 0x8000))
           ;

       /* start again at the beginning of copper list 1 */
       custom.cop1lc = p->coplist1hdr;
       custom.copjmp1 = 0;
   }
}
/* ====================================================================== */

static struct geometry modes[] =
{
    /* NTSC modes: !!! can't guarantee anything about overscan modes !!! */
    {
	"ntsc-lace", 1,
	BPC0_HIRES | BPC0_LACE,
	640, 400, 1,
	704, 480,
	0x71, 0x18,			/* diwstrt h,v */
	16				/* WORD aligned */
	},
    {
	"ntsc", 1,
	BPC0_HIRES,
	640, 200, 1,
	704, 240,
	0x71, 0x18,
	16				/* WORD aligned */
	},
    {
	"ntsc-lace-over", 1,
	BPC0_HIRES | BPC0_LACE,
	704, 480, 1,
	704, 480,
	0x71, 0x18,
	16				/* WORD aligned */
	},
    {
	"ntsc-over", 1,
	BPC0_HIRES,
	704, 240, 1,
	704, 240,
	0x71, 0x18,
	16				/* WORD aligned */
	},
    /* PAL modes. Warning:
     * 
     * PAL overscan causes problems on my machine because maximum diwstop_h
     * value seems to be ~0x1c2, rather than 0x1e0+ inferred by RKM 1.1
     * and 0x1d5 inferred by original `amicon.c' source. Is this a hardware
     * limitation of OCS/pal or 1084?. Or am I doing something stupid here?
     *
     * Included a couple of overscan modes that DO work on my machine,
     * although not particularly useful.
     */
    {
	"pal-lace", 1,
	BPC0_HIRES | BPC0_LACE,
	640, 512, 1,
	704, 592,
	0x71, 0x18,
	16				/* WORD aligned */
	},
    {
	"pal", 1,
	BPC0_HIRES,
	640, 256, 1,
	704, 296,
	0x71, 0x18,
	16				/* WORD aligned */
	},
    {
	"pal-lace-over", 1,
	BPC0_HIRES | BPC0_LACE,
	704, 592, 1,
	704, 582,
	0x5b, 0x18,
	16				/* WORD aligned */
	},
    {
	"pal-over", 1,
	BPC0_HIRES,
	704, 296, 1,
	704, 296,
	0x5b, 0x18,
	16				/* WORD aligned */
	},
    /* AGA modes */
    {
	/*
	 * A 640x480, 60Hz noninterlaced AGA mode. It would be nice to be
	 * able to have some of these values computed dynamically, but that
	 * requires more knowledge of AGA than I have. At the moment,
	 * the values make it centered on my 1960 monitor. -wjr
	 *
	 * For random reasons to do with the way arguments are parsed,
	 * these names can't start with a digit.
	 *
	 * Don't count on being able to reduce scr_width and scr_height
	 * and ending up with a smaller but well-formed screen - this
	 * doesn't seem to work well at the moment.
	 */
	"aga640x480", 0,
	BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
	640, 480, 1,
	640, 480,
	0x0041, 0x002b,					/* diwstrt h,v */
	64,						/* 64-bit aligned */
	BPC2_KILLEHB,					/* bplcon2 */
	BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
	    BPC3_BRDRBLNK | BPC3_EXTBLKEN,		/* bplcon3 */
	0x0071,						/* htotal */
	0x000c,						/* hsstrt */
	0x001c,						/* hsstop */
	0x0008,						/* hbstrt */
	0x001e,						/* hbstop */
	0x020c,						/* vtotal */
	0x0001,						/* vsstrt */
	0x0003,						/* vsstop */
	0x0000,						/* vbstrt */
	0x000f,						/* vbstop */
	0x0046,						/* hcenter */
	BMC0_HARDDIS | BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
	    BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,	/* beamcon0 */
	FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	},
	/* An 800x600 72Hz interlaced mode. */
	{
	"aga800x600", 0,
	BPC0_SHRES | BPC0_LACE | BPC0_ECSENA,		/* bplcon0 */
	896, 624, 1,					/* need rows%8 == 0 */
	896, 624,					/* (cols too) */
	0x0041, 0x001e,					/* diwstrt h,v */
	64,						/* 64-bit aligned */
	BPC2_KILLEHB,					/* bplcon2 */
	BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
	    BPC3_BRDRBLNK | BPC3_EXTBLKEN,		/* bplcon3 */
	0x0091,						/* htotal */
	0x000e,						/* hsstrt */
	0x001d,						/* hsstop */
	0x000a,						/* hbstrt */
	0x001e,						/* hbstop */
	0x0156,						/* vtotal */
	0x0001,						/* vsstrt */
	0x0003,						/* vsstop */
	0x0000,						/* vbstrt */
	0x000f,						/* vbstop */
	0x0050,						/* hcenter */
	BMC0_HARDDIS | BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
	    BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,	/* beamcon0 */
	FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	},
    /*
     * Additional AGA modes by Geert Uytterhoeven
     */
    {
	/*
	 * A 720x400, 70 Hz noninterlaced AGA mode (29.27 kHz)
	 */
	"aga720x400", 0,
	BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
	720, 400, 1,
	720, 400,
	0x0041, 0x0013,					/* diwstrt h,v */
	64,						/* 64-bit aligned */
	BPC2_KILLEHB,					/* bplcon2 */
	BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
	    BPC3_BRDRBLNK | BPC3_EXTBLKEN,		/* bplcon3 */
	0x0079,						/* htotal */
	0x000e,						/* hsstrt */
	0x0018,						/* hsstop */
	0x0001,						/* hbstrt */
	0x0021,						/* hbstop */
	0x01a2,						/* vtotal */
	0x0003,						/* vsstrt */
	0x0005,						/* vsstop */
	0x0000,						/* vbstrt */
	0x0012,						/* vbstop */
	0x0046,						/* hcenter */
	BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
	BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
	BMC0_VSYTRUE,	/* beamcon0 */
	FMODE_BPAGEM | FMODE_BPL32			/* fmode */
    },
    {
	/*
	 * A 640x400, 76 Hz noninterlaced AGA mode (31.89 kHz)
	 */
	"aga640x400", 0,
	BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
	640, 400, 1,
	640, 400,
	0x0041, 0x0015,					/* diwstrt h,v */
	64,						/* 64-bit aligned */
	BPC2_KILLEHB,					/* bplcon2 */
	BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
	    BPC3_BRDRBLNK | BPC3_EXTBLKEN,		/* bplcon3 */
	0x006f,						/* htotal */
	0x000d,						/* hsstrt */
	0x0018,						/* hsstop */
	0x0001,						/* hbstrt */
	0x0021,						/* hbstop */
	0x01a4,						/* vtotal */
	0x0003,						/* vsstrt */
	0x0005,						/* vsstop */
	0x0000,						/* vbstrt */
	0x0014,						/* vbstop */
	0x0046,						/* hcenter */
	BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
	BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
	BMC0_VSYTRUE,	/* beamcon0 */
	FMODE_BPAGEM | FMODE_BPL32			/* fmode */
    },
    {
	/*
	 * A 640x480, 64 Hz noninterlaced AGA mode (31.89 kHz)
	 */
	"aga640x480a", 0,
	BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
	640, 480, 1,
	640, 480,
	0x0041, 0x0015,					/* diwstrt h,v */
	64,						/* 64-bit aligned */
	BPC2_KILLEHB,					/* bplcon2 */
	BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
	    BPC3_BRDRBLNK | BPC3_EXTBLKEN,		/* bplcon3 */
	0x006f,						/* htotal */
	0x000e,						/* hsstrt */
	0x0018,						/* hsstop */
	0x0001,						/* hbstrt */
	0x0021,						/* hbstop */
	0x01f4,						/* vtotal */
	0x0003,						/* vsstrt */
	0x0005,						/* vsstop */
	0x0000,						/* vbstrt */
	0x0014,						/* vbstop */
	0x0046,						/* hcenter */
	BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
	BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
	BMC0_VSYTRUE,	/* beamcon0 */
	FMODE_BPAGEM | FMODE_BPL32			/* fmode */
    }
};

#define	NMODES	(sizeof(modes) / sizeof(struct geometry))

/* Some default modes */
#define	OCS_PAL_LOWEND_DEFMODE	5	/* PAL non-laced for 500/2000 */
#define	OCS_PAL_3000_DEFMODE	4	/* PAL laced for 3000 */
#define	OCS_NTSC_LOWEND_DEFMODE	1	/* NTSC non-laced for 500/2000 */
#define	OCS_NTSC_3000_DEFMODE	0	/* NTSC laced for 3000 */
#define	AGA_DEFMODE		8	/* 640x480 non-laced for AGA */

static int amicon_inverse = 0;
static int amicon_mode = -1;

void video_setup (char *options, int *ints)
{
    char *this_opt;
    int i;
    
    printk ("video_setup: options are %s\n", options);
    
    if (!options || !*options) {
        return;
	}
    
    for (this_opt = strtok(options,",");
	 this_opt;
	 this_opt = strtok(NULL,",")) {
        if (!strcmp (this_opt, "inverse")) {
            amicon_inverse = 1;
	    }
        else {
	    for (i = 0; i < NMODES; i++) {
		if (!strcmp(this_opt, modes[i].modename)) {
		    amicon_mode = i;
		    break;
		    }
		}
	    }
	}
    }

static long amicon_init(struct condata *conp, long kmem_start)
{
   int unit = conp - vc_cons;
   register struct display *p = &disp[unit];
   int mode = amicon_mode;
   int inverse_video = amicon_inverse;
   ulong  model;

   model = boot_info.bi_un.bi_ami.model;

   if (mode == -1) { 
       if (boot_info.bi_amiga.chipset == CS_AGA) {
	   mode = AGA_DEFMODE;
	   }
       else {
	   if (model == AMI_3000) {
	       mode = (boot_info.bi_un.bi_ami.vblank == 50) ?
		   OCS_PAL_3000_DEFMODE : OCS_NTSC_3000_DEFMODE;
	       }
	   else {
	       mode = (boot_info.bi_un.bi_ami.vblank == 50) ?
		   OCS_PAL_LOWEND_DEFMODE : OCS_NTSC_LOWEND_DEFMODE;
	       }
	   }
       }

   init_vblank();
   display_init(p, &modes[mode], inverse_video);

   conp->vc_cols = p->scr_width  / p->fontwidth;
   conp->vc_rows = p->scr_height / p->fontheight;

   if(!mach_add_isr(IRQ_AMIGA_VERTB, (isrfunc)amicon_interrupt, 0, NULL))
           panic("Couldn't add vblank interrupt");

#ifdef DEBUG
   printk("amicon_init: fired up mode %s\n", modes[mode].modename);
   printk("p->diwstrt = (%x, %x), diwstop = (%x, %x)\n",
	  p->diwstrt_h, p->diwstrt_v, p->diwstop_h, p->diwstop_v);
#endif
   return kmem_start;
}

static int amicon_deinit (struct condata *conp)
{
    return 0;
}

/* ====================================================================== */

/* Those of a delicate disposition might like to skip the next couple of
 * pages.
 *
 * These functions are drop in replacements for memmove and
 * memset(_, 0, _). However their five instances add at least a kilobyte
 * to the object file. You have been warned.
 *
 * Not a great fan of assembler for the sake of it, but I think
 * that these routines are at least 10 times faster than their C
 * equivalents for large blits, and thats important to the lowest level of
 * a graphics driver. Question is whether some scheme with the blitter
 * would be faster. I suspect not for simple text system - not much
 * asynchronisity.
 *
 * Code is very simple, just gruesome expansion. Basic strategy is to
 * increase data moved/cleared at each step to 16 bytes to reduce
 * instruction per data move overhead. movem might be faster still
 * For more than 15 bytes, we try to align the write direction on a
 * longword boundary to get maximum speed. This is even more gruesome.
 * Unaligned read/write used requires 68020+ - think this is a problem?
 *
 * Sorry!
*/


#if defined (__GNUC__) && defined (__mc68000__)

static __inline__ void *mymemclear_small(void * s, size_t count)
{
__asm__("movel  %1,d0 ; beq 4f ; movel %0,a0 ; movel #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb d1,a0@+\n\t"
     "1: lsrl   #1,d0 ; jcc 2f ; movew d1,a0@+\n\t"
     "2: lsrl   #1,d0 ; jcc 3f ; movel d1,a0@+\n\t"
     "3: lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
     "5: movel d1,a0@+; movel d1,a0@+ ; movel d1,a0@+ ; movel d1,a0@+\n\t"
        "dbra d0,5b\n\t"
     "6:"
      :
           : "g" (s), "g" (count)
           : "a0", "d0", "d1");

   return(0);              
}

static __inline__ void *mymemclear(void * s, size_t count)
{
    if (count < 16) 
__asm__("movel  %1,d0 ; beq 4f ; movel %0,a0   ; movel #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb d1,a0@+\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew d1,a0@+\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel d1,a0@+\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
      "4:"
   :
        : "g" (s), "g" (count)
   : "a0", "d0", "d1");
    else
__asm__("movel  %1,d0 ; movel  %0,a0 ; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; clrb a0@+ ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "clrw   a0@+  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "clrw   a0@+  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f ; moveql #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel  d1,a0@+\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
     "5: movel d1,a0@+; movel d1,a0@+ ; movel d1,a0@+ ; movel d1,a0@+\n\t"
         "dbra d0,5b; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew d1,a0@+\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb d1,a0@+\n\t"
     "8:"
   :
        : "g" (s), "g" (count)
   : "a0", "d0", "d1");
   return(0);              
}

static __inline__ void *mymemmove(void * d, void * s, size_t count)
{
    if (d < s) {
        if (count < 16) {
__asm__("movel  %2,d0 ; beq 4f ; movel %0,a0   ; movel %1, a1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb a1@+,a0@+\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew a1@+,a0@+\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel a1@+,a0@+\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel a1@+,a0@+ ; movel a1@+,a0@+\n\t"
      "4:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0");
       } else {
__asm__("movel  %2,d0 ; movel  %0,a0 ; movel %1,a1; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; moveb a1@+,a0@+ ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "movew  a1@+,a0@+  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "movew  a1@+,a0@+  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel a1@+,a0@+\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel a1@+,a0@+ ; movel a1@+,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
          "5: movel a1@+,a0@+;movel a1@+,a0@+\n\t"
             "movel a1@+,a0@+;movel a1@+,a0@+\n\t"
             "dbra d0,5b ; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew a1@+,a0@+\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb a1@+,a0@+\n\t"
     "8:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0", "d1");
      }
   } else {
    if (count < 16) {
__asm__("movel  %2,d0 ; beq 4f ; movel %0,a0   ; movel %1, a1\n\t"
        "addal  d0,a0 ; addal d0,a1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb a1@-,a0@-\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew a1@-,a0@-\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel a1@-,a0@-\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel a1@-,a0@- ; movel a1@-,a0@-\n\t"
      "4:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0");
      } else {
__asm__("movel  %2,d0 ; movel  %0,a0 ; movel %1,a1\n\t"
        "addal  d0,a0 ; addal d0,a1; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; moveb a1@-,a0@- ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "movew  a1@-,a0@-  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "movew  a1@-,a0@-  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel a1@-,a0@-\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel a1@-,a0@- ; movel a1@-,a0@-\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
          "5: movel a1@-,a0@-;movel a1@-,a0@-\n\t"
             "movel a1@-,a0@-;movel a1@-,a0@-\n\t"
             "dbra d0,5b ; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew a1@-,a0@-\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb a1@-,a0@-\n\t"
     "8:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0", "d1");
      }
   }
   return(0);              
}
#else
#define mymemmove(d, s, count)       memmove((d),(s),(count))
#define mymemclear(d, count)         memset((d), 0 ,(count))
#define mymemclear_small(d, count)   memset((d), 0 ,(count))
#endif

/* ====================================================================== */

/* amicon_XXX routines - interface used by the world
 *
 * This system is now divided into two levels because of complications
 * caused by hardware scrolling. Top level functions:
 *
 *    amicon_bmove(), amicon_clear(), amicon_putc()
 *
 * handles y values in range [0, scr_height-1] that correspond to real
 * screen positions. y_wrap shift means that first line of bitmap may be
 * anywhere on this display. These functions convert lineoffsets to
 * bitmap offsets and deal with the wrap-around case by splitting blits.
 *
 *    amicon_bmove_physical_8()   -- These functions fast implementations
 *    amicon_clear_physical_8()   -- of original amicon_XXX fns.
 *    amicon_putc_physical_8()    -- (fontwidth  != 8) may be added later
 *
 * WARNING:
 *
 * At the moment amicon_putc() cannot blit across vertical wrap boundary
 * Implies should only really hardware scroll in rows. Only reason for
 * restriction is simplicity & efficiency at the moment.
 *
*/

static __inline__ int real_y_8(register struct display *p, int y)
{
   int y_bottom_size = p->y_wrap >> 3;
   int y_top_size    = (p->scr_height - p->y_wrap) >> 3;

   return( (y < y_top_size) ? (y + y_bottom_size) : (y-y_top_size) );
}

static void amicon_bmove_physical_8 (register struct display *p,
                                    int sy, int sx, int dy, int dx,
                                    int height, int width)
{

    if (sx == 0 && sy == 0 && width == p->bytes_per_row) {
        register ushort i;
        for (i = 0 ; i < p->scr_depth ; i++)
            mymemmove(p->bitplane[i] + dy * width * p->fontheight,
                      p->bitplane[i] + sy * width * p->fontheight,
                      width * height * p->fontheight);
    } else {
        register ushort rows, i;
        register u_char *src;
        register u_char *dst;
        register ushort bytes      = p->bytes_per_row;
        register ushort linesize   = bytes * p->fontheight;
        
        if (dy <= sy) {
            for (i = 0 ; i < p->scr_depth ; i++) {
                src  = p->bitplane[i] + sy * linesize + sx;
                dst  = p->bitplane[i] + dy * linesize + dx;
                for (rows = height*p->fontheight ; rows-- ;) {
                    mymemmove(dst, src, width);
                    src += bytes;
                    dst += bytes;
                }
            }
        } else {
            for (i = 0 ; i < p->scr_depth ; i++) {
                src  = p->bitplane[i] + (sy+height) * linesize - bytes + sx;
                dst  = p->bitplane[i] + (dy+height) * linesize - bytes + dx;

                for (rows = height*p->fontheight ; rows-- ;) {
                    mymemmove(dst, src, width);
                    src -= bytes;
                    dst -= bytes;
                }
            }
        }
    }
}

static int amicon_bmove (struct condata *conp,
          int sy, int sx, int dy, int dx,
          int height, int width)
{
    int unit = conp - vc_cons;
    struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        /* Split blits that cross physical y_wrap case.
         * Pathological case involves 4 blits, better to use recursive
         * code rather than unrolled case
        */

        ushort y_break = (p->scr_height - p->y_wrap) >> 3;

        if (sy < y_break && (sy+height-1) >= y_break) {
            ushort b = y_break - sy;
            if (dy < sy) {  /* Avoid trashing self */
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
            } else {
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
            }
            return(0);
        }
       
        if (dy < y_break && (dy+height-1) >= y_break) {
            ushort b = y_break - dy;
            if (dy < sy) {  /* Avoid trashing self */
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
            } else {
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
            }
            return(0);
        }
        amicon_bmove_physical_8(p, real_y_8(p,sy),   sx,
                                   real_y_8(p,dy),   dx,
                                   height,           width);
    } else {
        /* No support (fontwidth != 8) yet */
    }

    return(0);
}

static void amicon_clear_physical_8 (register struct display *p,
                                     int sy, int sx, int height, int width)
{
    ulong offset;
    register u_char *start;
    register ushort i, rows;
    register ushort bytes = p->bytes_per_row;
    register ulong size;

    if (sx == 0 && width == bytes) {
        offset = sy     * bytes * p->fontheight;
        size   = height * bytes * p->fontheight;
        
        for (i = 0 ; i < p->scr_depth ; i++)
            mymemclear(p->bitplane[i]+offset, size);
    } else {
        offset = (sy * bytes * p->fontheight) + sx;
        
        for (i = 0 ; i < p->scr_depth ; i++) {
            start = p->bitplane[i]+offset;
                
            for (rows = height*p->fontheight; rows-- ; start += bytes)
                mymemclear_small(start, width);
        }
    }
}

static int amicon_clear (struct condata *conp,
                         int sy, int sx, int height, int width)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        /* Split blits that cross physical y_wrap boundary */

        ushort y_break = (p->scr_height - p->y_wrap) >> 3;

        if (sy < y_break && (sy+height-1) >= y_break) {
            ushort b = y_break - sy;
            amicon_clear_physical_8(p, real_y_8(p,sy),   sx,
                                       b,                width);
            amicon_clear_physical_8(p, real_y_8(p,sy+b), sx,
                                       height-b,         width);

        } else
            amicon_clear_physical_8(p, real_y_8(p,sy),   sx,
                                       height,           width);
    } else {
        /* No support (fontwidth != 8) yet */
    }

    return(0);
}

static void amicon_putc_physical_8 (struct display *p,
                                    int c, int y, int x, int mode)
{
    register u_char   *dest;
    register u_char   *cdat;
    register ushort   rows;
    register ushort   i;
    u_short           bytes;
    ulong             offset;
        
    if ((c >= 32) && (c < 127))
   /* lower half of charset -- normal ASCII */
   c -=  32;
    else if ((c >= 160) && (c < 255))
   /* upper half of charset -- special characters */
   c -= 64;
    else
   /* not defined in charset -- non-printable character */
   return;

    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;
    dest  = p->bitplane[0] + offset;
    cdat  = p->fontdata + (c * p->fontheight);

    /* Arbitary decision to write chars as colour 1 for the moment,
     * regardless of number of bitplanes. Later will write/clear
     * based on colour bitmask.
     */

    for (rows = p->fontheight ; rows-- ; dest += bytes)
        *dest = *cdat++;

    for (i = 1 ; i < p->scr_depth ; i++) {
        dest = p->bitplane[i] + offset;
        for (rows = p->fontheight ; rows-- ; dest += bytes)
            *dest = (u_char)0;
    }
}

static int amicon_putc (struct condata *conp,
                        int c, int y, int x, int mode)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        amicon_putc_physical_8(p, c, real_y_8(p, y) , x, mode);
    } else {
        /* Currently no support for (fontwidth != 8) */
    }

    return(0);
}

static void amicon_putcs_physical_8 (struct display *p, const char *s,
                                     int count, int y, int x, int mode)
{
    register u_char   *dest, *cdat;
    register ushort   rows, bytes;
    register u_char   c;
    u_char   *dest0;
    ulong    offset;
    ushort   i;

    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;

    for (i = 1 ; i < p->scr_depth ; i++) {
        dest = p->bitplane[i] + offset;
        for (rows = p->fontheight ; rows-- ; dest += bytes)
            mymemclear_small(dest, count);
    }

    dest0  = p->bitplane[0] + offset;

#if defined (__GNUC__) && defined (__mc68000__)
    if (p->fontheight == 8) {
        /* Ultra special support for 8x8 fonts, as its nice and easy
         * and the console driver seems to spend most of its time in
         * here. Would take another 4 instructions to support,
         * (fontheight != 8) but I have an inbuilt suspicion of mulu's.
         * A hangover from my 68000 days, I guess.
         *
         * Both code and register use mirrors the C code directly below:
         *
         * a0 -- dest & dest0  [uses stack] 
         * a1 -- cdat     ; a2 -- s      ; a3 -- fontdata
         * d0 -- scratch  ; d1 -- count  ; d2 -- bytes
         * 
         */
        __asm__("movel  %1,d1   ; beq 5f      ; subqw #1,d1\n\t"
                "movel  %0,a0   ; movel %2,a2 ; movel %3,a3;movel %4,d2\n\t"
             "1: moveb  a2@+,d0 ; extw d0\n\t"
                "cmpb   #32,d0  ; jlt 2f ; cmpb #127,d0 ; jge 2f\n\t"
                   "subb   #32,d0  ; jra 3f\n\t"
             "2: cmpb   #160,d0 ; jlt 1b ; cmpb #255,d0 ; jeq 1b\n\t"
                   "subb   #64,d0\n\t"
             "3: lea    a3@(d0:w:8),a1 ; movel a0,sp@-\n\t"
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 0 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 1 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 2 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 3 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 4 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 5 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 6 */
               " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 7 */
                "movel  sp@+,a0 ; lea a0@(1),a0   ; dbra d1,1b\n\t"
             "5:"
                :
                : "g" (dest0), "g" (count), "g" (s),
                  "g" (p->fontdata), "g" ((ulong) bytes)
                : "a0", "a1", "a2", "a3", "d0", "d1", "d2");
        return;
    }
#endif

    while (count--) {
        c    = *(s++); dest = dest0++;
            
        /* Arbitary decision to write chars as colour 1 for the moment,
         * regardless of number of bitplanes. Later will write/clear
         * based on colour bitmask.
         */

        if ((c >= 32) && (c < 127))
            /* lower half of charset -- normal ASCII */
            c -=  32;
        else if ((c >= 160) && (c < 255))
            /* upper half of charset -- special characters */
            c -= 64;
        else
            continue;

        cdat  = p->fontdata + (c * p->fontheight);
        for (rows = p->fontheight ; rows-- ; dest += bytes)
            *dest = *cdat++;
    }
}

static int amicon_putcs (struct condata *conp, const char *s,
                         int count, int y, int x, int mode)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        amicon_putcs_physical_8(p, s, count, real_y_8(p, y) , x, mode);
    } else {
        /* Currently no support for (fontwidth != 8) */
    }

    return(0);
}

static int amicon_cursor (struct condata *conp,
           int mode)
{
   int unit = conp - vc_cons;
   struct display *p = &disp[unit];

   if (mode == CM_ERASE) {
       /* *(long *)p->cursor = 0; */
   } else {
       p->cursor_latch = 0;
       p->cursor_x     = (ushort) conp->vc_x;
       p->cursor_y     = (ushort) conp->vc_y;
       p->cursor_latch = 1;
   }

    return 0;
}

static int amicon_scroll (struct condata *conp,
                          int t, int b, int dir, int count)
{
    int unit                   = conp - vc_cons;
    register struct display *p = &disp[unit];

    amicon_cursor (conp, CM_ERASE);

    switch (dir) {
    case SM_UP:
        if (t == 0 && b == conp->vc_rows) {
            /* Can hardware scroll */

            p->scroll_latch = 0;
            if (count > conp->vc_rows)         /* Maximum realistic size */
                count = conp->vc_rows;
            
            p->y_wrap += p->fontheight * count;
            if (p->y_wrap > p->scr_height)     /* Deal with wrap */
                p->y_wrap -= p->scr_height;

            p->scroll_latch = 1;
        } else
            amicon_bmove(conp,t + count, 0, t, 0,
                         (b-t-count), p->bytes_per_row);

   amicon_clear(conp,b-count,0,count,p->bytes_per_row);
        break;

    case SM_DOWN:
        if (t == 0 && b == conp->vc_rows) {
            /* Can hardware scroll */

            p->scroll_latch = 0;
            if (count > conp->vc_rows)         /* Maximum realistic size */
                count = conp->vc_rows;

            if (((short)p->y_wrap - p->fontheight * count) < 0)
                p->y_wrap += p->scr_height;    /* y_wrap is U_short */

            p->y_wrap -= p->fontheight * count;
            p->scroll_latch = 1;
        } else
            amicon_bmove(conp,t,0,t+count,0,
                         (b-t-count), p->bytes_per_row);

        /* Fixed bmove() should end Arno's frustration with copying?
         * Confusius say:
         *    Man who copies in wrong direction, end up with trashed data
        */

   amicon_clear(conp, t, 0, count, p->bytes_per_row);
        break;

    case SM_LEFT:
    case SM_RIGHT:
        /* XXX Nothing here yet XXX */

        break;
    }

    return 0;
}

static int amicon_switch (struct condata *conp)
{
    return 0;
}

static int amicon_blank (int blank)
{
    if (blank)
	do_blank = 1;
    else
	do_unblank = 1;

    return 0;
}


/*
 * The console "switch" structure for the Amiga native console
 */

struct consw ami_con =
{
    amicon_init, amicon_deinit, amicon_clear, amicon_putc, amicon_putcs,
    amicon_cursor, amicon_scroll, amicon_bmove, amicon_switch, amicon_blank
};

