/*
 * linux/amiga/amikeyb.c
 *
 * Amiga Keyboard driver for 680x0 Linux
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

/*
 * Amiga support by Hamish Macdonald
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/keyboard.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>

#include <linux/amigatypes.h>
#include <linux/amigaints.h>
#include <linux/amigahw.h>

extern void process_scancode (int);

#define AMIKEY_CAPS	(0x62)
#define BREAK_MASK	(0x80)

unsigned short amikey_map[NR_KEYMAPS][NR_KEYS] = {
    { /* no modifiers */
/* 00-03 */	'`',	'1',	'2',	'3',
/* 04-07 */	'4',	'5',	'6',	'7',
/* 08-0b */	'8',	'9',	'0',	'-',
/* 0c-0f */	'=',	'\\',	K_HOLE,	K_P0,
/* 10-13 */	'q',	'w',	'e',	'r',
/* 14-17 */	't',	'y',	'u',	'i',
/* 18-1b */	'o',	'p',	'[',	']',
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	'a',	's',	'd',	'f',
/* 24-27 */	'g',	'h',	'j',	'k',
/* 28-2b */	'l',	';',	'\'',	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	'z',	'x',	'c',
/* 34-37 */	'v',	'b',	'n',	'm',
/* 38-3b */	',',	'.',	'/',	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	' ',	'?'|0x40,	'I'&0x1f,	K_PENTER,
/* 44-47 */	K_ENTER,	'['&0x1f,	'?'|0x40,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_F1,	K_F2,	K_F3,	K_F4,
/* 54-57 */	K_F5,	K_F6,	K_F7,	K_F8,
/* 58-5b */	K_F9,	K_F10,	K_NUM,	K_HOLD,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_F11,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift */
/* 00-03 */	'~',	'!',	'@',	'#',
/* 04-07 */	'$',	'%',	'^',	'&',
/* 08-0b */	'*',	'(',	')',	'_',
/* 0c-0f */	'+',	'|',	K_HOLE,	K_P0,
/* 10-13 */	'Q',	'W',	'E',	'R',
/* 14-17 */	'T',	'Y',	'U',	'I',
/* 18-1b */	'O',	'P',	'{',	'}',
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	'A',	'S',	'D',	'F',
/* 24-27 */	'G',	'H',	'J',	'K',
/* 28-2b */	'L',	':',	'"',	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	'Z',	'X',	'C',
/* 34-37 */	'V',	'B',	'N',	'M',
/* 38-3b */	'<',	'>',	'?',	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	' ',	'?'|0x40,	'I'&0x1f,	K_PENTER,
/* 44-47 */	K_ENTER,	'['&0x1f,	'?'|0x40,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_F11,	K_F12,	K_F13,	K_F14,
/* 54-57 */	K_F15,	K_F16,	K_F17,	K_F18,
/* 58-5b */	K_F19,	K_F20,	K_NUM,	K_SH_MEM,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_F11,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	'@',	K_HOLE,
/* 04-07 */	'$',	K_HOLE,	K_HOLE,	'{',
/* 08-0b */	'[',	']',	'}',	'\\',
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	'~',
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K(KT_CONS,12),	K(KT_CONS,13),	K(KT_CONS,14),	K(KT_CONS,15),
/* 54-57 */	K(KT_CONS,16),	K(KT_CONS,17),	K(KT_CONS,18),	K(KT_CONS,19),
/* 58-5b */	K(KT_CONS,20),	K(KT_CONS,21),	K_NUM,	K_SH_REGS,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K(KT_CONS,22),
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control */
/* 00-03 */	'@'&0x1f,	K_HOLE,	'@'&0x1f,	'['&0x1f,
/* 04-07 */	'\\'&0x1f,	']'&0x1f,	'^'&0x1f,	'_'&0x1f,
/* 08-0b */	'?'|0x40,	K_HOLE,	K_HOLE,	'_'&0x1f,
/* 0c-0f */	K_HOLE,	'\\'&0x1f,	K_HOLE,	K_P0,
/* 10-13 */	'Q'&0x1f,	'W'&0x1f,	'E'&0x1f,	'R'&0x1f,
/* 14-17 */	'T'&0x1f,	'Y'&0x1f,	'U'&0x1f,	'I'&0x1f,
/* 18-1b */	'O'&0x1f,	'P'&0x1f,	'['&0x1f,	']'&0x1f,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	'A'&0x1f,	'S'&0x1f,	'D'&0x1f,	'F'&0x1f,
/* 24-27 */	'G'&0x1f,	'H'&0x1f,	'J'&0x1f,	'K'&0x1f,
/* 28-2b */	'L'&0x1f,	K_HOLE,	'G'&0x1f,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	'Z'&0x1f,	'X'&0x1f,	'C'&0x1f,
/* 34-37 */	'V'&0x1f,	'B'&0x1f,	'N'&0x1f,	'M'&0x1f,
/* 38-3b */	K_HOLE,	K_HOLE,	'?'|0x40,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	'@'&0x1f,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_F1,	K_F2,	K_F3,	K_F4,
/* 54-57 */	K_F5,	K_F6,	K_F7,	K_F8,
/* 58-5b */	K_F9,	K_F10,	K_NUM,	K_SH_STAT,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_F11,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_BOOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* alt */
/* 00-03 */	K(KT_META,'`'),	K(KT_META,'1'),	K(KT_META,'2'),	K(KT_META,'3'),
/* 04-07 */	K(KT_META,'4'),	K(KT_META,'5'),	K(KT_META,'6'),	K(KT_META,'7'),
/* 08-0b */	K(KT_META,'8'),	K(KT_META,'9'),	K(KT_META,'0'),	K(KT_META,'-'),
/* 0c-0f */	K(KT_META,'='),	K(KT_META,'\\'),	K_HOLE,	K_ASC0,
/* 10-13 */	K(KT_META,'q'),	K(KT_META,'w'),	K(KT_META,'e'),	K(KT_META,'r'),
/* 14-17 */	K(KT_META,'t'),	K(KT_META,'y'),	K(KT_META,'u'),	K(KT_META,'i'),
/* 18-1b */	K(KT_META,'o'),	K(KT_META,'p'),	K(KT_META,'['),	K(KT_META,']'),
/* 1c-1f */	K_HOLE,	K_ASC1,	K_ASC2,	K_ASC3,
/* 20-23 */	K(KT_META,'a'),	K(KT_META,'s'),	K(KT_META,'d'),	K(KT_META,'f'),
/* 24-27 */	K(KT_META,'g'),	K(KT_META,'h'),	K(KT_META,'j'),	K(KT_META,'k'),
/* 28-2b */	K(KT_META,'l'),	K(KT_META,';'),	K(KT_META,'\''),	K_HOLE,
/* 2c-2f */	K_HOLE,	K_ASC4,	K_ASC5,	K_ASC6,
/* 30-33 */	K_HOLE,	K(KT_META,'z'),	K(KT_META,'x'),	K(KT_META,'c'),
/* 34-37 */	K(KT_META,'v'),	K(KT_META,'b'),	K(KT_META,'n'),	K(KT_META,'m'),
/* 38-3b */	K(KT_META,','),	K(KT_META,'.'),	K(KT_META,'/'),	K_HOLE,
/* 3c-3f */	K_PDOT,	K_ASC7,	K_ASC8,	K_ASC9,
/* 40-43 */	K(KT_META,' '),	K(KT_META,'?'|0x40),	K(KT_META,'I'&0x1f),	K_PENTER,
/* 44-47 */	K(KT_META,'M'&0x1f),	K(KT_META,'['&0x1f),	K(KT_META,'?'|0x40),	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K(KT_CONS,0),	K(KT_CONS,1),	K(KT_CONS,2),	K(KT_CONS,3),
/* 54-57 */	K(KT_CONS,4),	K(KT_CONS,5),	K(KT_CONS,6),	K(KT_CONS,7),
/* 58-5b */	K(KT_CONS,8),	K(KT_CONS,9),	K_NUM,	K_HOLD,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K(KT_CONS,10),
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift alt */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* alt altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift alt altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control alt */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_BOOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control alt */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control alt altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control alt altgr */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K_HOLE,	K_P0,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_HOLE,	K_P1,	K_P2,	K_P3,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_P4,	K_P5,	K_P6,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 38-3b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 3c-3f */	K_PDOT,	K_P7,	K_P8,	K_P9,
/* 40-43 */	K_HOLE,	'?'|0x40,	K_HOLE,	K_PENTER,
/* 44-47 */	K_ENTER,	K_HOLE,	K_HOLE,	K_HOLE,
/* 48-4b */	K_HOLE,	K_HOLE,	K_PMINUS,	K_HOLE,
/* 4c-4f */	K_UP,	K_DOWN,	K_RIGHT,	K_LEFT,
/* 50-53 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_NUM,	K_HOLE,
/* 5c-5f */	K_PSLASH,	K_PSTAR,	K_PPLUS,	K_HOLE,
/* 60-63 */	K_SHIFT,	K_SHIFT,	K_CAPS,	K_CTRL,
/* 64-67 */	K_ALT,	K_ALTGR,	K_HOLE,	K_HOLE,
/* 68-6b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 6c-6f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 70-73 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    },
};

static unsigned char rep_scancode;
static void amikeyb_rep (unsigned long ignore);
static struct timer_list amikeyb_rep_timer = {NULL, NULL, 0, 0, amikeyb_rep};

static void amikeyb_rep (unsigned long ignore)
{
	unsigned char scancode = rep_scancode;
	amikeyb_rep_timer.expires = HZ/20;
	add_timer(&amikeyb_rep_timer);
	process_scancode (scancode);
}

static void keyboard_interrupt(struct intframe *fp, void *data)
{
    unsigned char scancode, break_flag;

    /* get and invert scancode (keyboard is active low) */
    scancode = ~ciaa.sdr;

    /* switch SP pin to output for handshake */
    ciaa.cra |= 0x40;
    /* wait until 85 us have expired */
    udelay(85);
    /* switch CIA serial port to input mode */
    ciaa.cra &= ~0x40;

    /* rotate scan code to get up/down bit in proper position */
    __asm__ __volatile__ ("rorb #1,%0" : "=g" (scancode) : "0" (scancode));

    /*
     * do machine independent keyboard processing of "normalized" scancode
     * A "normalized" scancode is one that an IBM PC might generate
     * Check make/break first
     */
    break_flag = scancode & BREAK_MASK;
    scancode &= (unsigned char )~BREAK_MASK;

    if (scancode == AMIKEY_CAPS) {
	    /* if the key is CAPS, fake a press/release. */
	    process_scancode (AMIKEY_CAPS);
	    process_scancode (BREAK_MASK | AMIKEY_CAPS);
    } else {
	    /* handle repeat */
	    if (break_flag) {
		    del_timer(&amikeyb_rep_timer);
		    rep_scancode = 0;
	    } else {
		    del_timer(&amikeyb_rep_timer);
		    rep_scancode = break_flag | scancode;
		    amikeyb_rep_timer.expires = (HZ*2)/3;
		    add_timer(&amikeyb_rep_timer);
	    }
	    process_scancode (break_flag | scancode);
    }

    return;
}

unsigned long amiga_keyb_init (unsigned long mem_start)
{
    /* setup key map */
    key_map = amikey_map;

    /*
     * Initialize serial data direction.
     */
    ciaa.cra &= ~0x41;	     /* serial data in, turn off TA */

    /*
     * arrange for processing of keyboard interrupt
     */
    add_isr (IRQ_AMIGA_CIAA_SP, keyboard_interrupt, 0, NULL);

    return mem_start;
}
